! ***********************************************************************
!
!   Copyright (C) 2010-2019  Bill Paxton & The MESA Team
!
!   this file is part of mesa.
!
!   mesa is free software; you can redistribute it and/or modify
!   it under the terms of the gnu general library public license as published
!   by the free software foundation; either version 2 of the license, or
!   (at your option) any later version.
!
!   mesa is distributed in the hope that it will be useful, 
!   but without any warranty; without even the implied warranty of
!   merchantability or fitness for a particular purpose.  see the
!   gnu library general public license for more details.
!
!   you should have received a copy of the gnu library general public license
!   along with this software; if not, write to the free software
!   foundation, inc., 59 temple place, suite 330, boston, ma 02111-1307 usa
!
! ***********************************************************************
 
      module run_star_extras

      use star_lib
      use star_def
      use const_def
      use math_lib
      
      implicit none

      logical :: has_flashed
      
      ! these routines are called by the standard run_star check_model
      contains

      subroutine extra_envelope_mixing(id, ierr)
         use chem_def, only: ih1
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         integer :: k
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return

         if (s% L_nuc_burn_total .gt. s% x_ctrl(1)) then
            do k=1, s% nz
               if (s% xa(s% net_iso(ih1), k) .gt. 0.5) then
                  s% D_mix(k) = s% D_mix(k) + s% x_ctrl(2)
               end if
            end do
         end if

      end subroutine extra_envelope_mixing
      
      subroutine extras_controls(id, ierr)
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         
         ! this is the place to set any procedure pointers you want to change
         ! e.g., other_wind, other_mixing, other_energy  (see star_data.inc)
         s% other_D_mix => extra_envelope_mixing

         ! the extras functions in this file will not be called
         ! unless you set their function pointers as done below.
         ! otherwise we use a null_ version which does nothing (except warn).

         s% extras_startup => extras_startup
         s% extras_start_step => extras_start_step
         s% extras_check_model => extras_check_model
         s% extras_finish_step => extras_finish_step
         s% extras_after_evolve => extras_after_evolve
         s% how_many_extra_history_columns => how_many_extra_history_columns
         s% data_for_extra_history_columns => data_for_extra_history_columns
         s% how_many_extra_profile_columns => how_many_extra_profile_columns
         s% data_for_extra_profile_columns => data_for_extra_profile_columns  

         s% how_many_extra_history_header_items => how_many_extra_history_header_items
         s% data_for_extra_history_header_items => data_for_extra_history_header_items
         s% how_many_extra_profile_header_items => how_many_extra_profile_header_items
         s% data_for_extra_profile_header_items => data_for_extra_profile_header_items

      end subroutine extras_controls
      
      
      subroutine extras_startup(id, restart, ierr)
         integer, intent(in) :: id
         logical, intent(in) :: restart
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         has_flashed = .false.
      end subroutine extras_startup
      

      integer function extras_start_step(id)
         integer, intent(in) :: id
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_start_step = 0
      end function extras_start_step


      ! returns either keep_going, retry, or terminate.
      integer function extras_check_model(id)
         integer, intent(in) :: id
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_check_model = keep_going         

         ! if you want to check multiple conditions, it can be useful
         ! to set a different termination code depending on which
         ! condition was triggered.  MESA provides 9 customizeable
         ! termination codes, named t_xtra1 .. t_xtra9.  You can
         ! customize the messages that will be printed upon exit by
         ! setting the corresponding termination_code_str value.
         ! termination_code_str(t_xtra1) = 'my termination condition'

         ! by default, indicate where (in the code) MESA terminated
         if (extras_check_model == terminate) s% termination_code = t_extras_check_model
      end function extras_check_model


      integer function how_many_extra_history_columns(id)
         integer, intent(in) :: id
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_history_columns = 2
      end function how_many_extra_history_columns
      
      
      subroutine data_for_extra_history_columns(id, n, names, vals, ierr)
         use chem_def, only: ih1, ihe4, ili7
         integer, intent(in) :: id, n
         character (len=maxlen_history_column_name) :: names(n)
         real(dp) :: vals(n)
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         integer :: k
         real(dp) :: max_mach
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         
         ! note: do NOT add the extras names to history_columns.list
         ! the history_columns.list is only for the built-in history column options.
         ! it must not include the new column names you are adding here.
         names(1) = 'A_Li'
         vals(1) = log10((s% xa(s% net_iso(ili7),1)/7d0)/(s% xa(s% net_iso(ih1),1)/1d0)) + 12d0

         names(2) = 'max_mlt_vc_div_csound' ! in He core
         max_mach = 0
         do k = 1, s% nz
            if (s% xa(s% net_iso(ihe4),k) .gt. 0.5) then
               max_mach = max(s% mlt_vc(k)/s% csound(k), max_mach)
            end if
         end do
         vals(2) = max_mach

      end subroutine data_for_extra_history_columns

      
      integer function how_many_extra_profile_columns(id)
         integer, intent(in) :: id
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_profile_columns = 3
      end function how_many_extra_profile_columns
      
      
      subroutine data_for_extra_profile_columns(id, n, nz, names, vals, ierr)
         use star_def, only: star_info, maxlen_profile_column_name
         use const_def, only: dp
         use net_lib, only: net_work_size
         integer, intent(in) :: id, n, nz
         character (len=maxlen_profile_column_name) :: names(n)
         real(dp) :: vals(nz,n)
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         integer :: k, op_err, net_lwork
         logical :: okay
         real(dp) :: sig, D_th
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         net_lwork = net_work_size(s% net_handle, ierr)        
         names(1) = 'log_rate_r_be7_wk_li7'
         names(2) = 'log_rate_r_li7_pa_he4'
         okay = .true.
         !$OMP PARALLEL DO PRIVATE(k,op_err) 
         do k = 1, s% nz
            if (.not. okay) cycle
            op_err = 0
            call do1_net( &
                 s, k, s% species, &
                 s% num_reactions, net_lwork, &
                 n, nz, vals, op_err)
            if (op_err /= 0) okay = .false.
         end do
         !$OMP END PARALLEL DO

         names(3) = 'D_thermal'
         do k=1, nz
            sig = 4d0*crad*clight*s% T(k)**3 / (s% Rho(k)*s% opacity(k))
            D_th = sig / (s% cp(k) * s% Rho(k))
            vals(k,3) = D_th
         end do

      end subroutine data_for_extra_profile_columns


      integer function how_many_extra_history_header_items(id)
         integer, intent(in) :: id
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_history_header_items = 0
      end function how_many_extra_history_header_items


      subroutine data_for_extra_history_header_items(id, n, names, vals, ierr)
         integer, intent(in) :: id, n
         character (len=maxlen_history_column_name) :: names(n)
         real(dp) :: vals(n)
         type(star_info), pointer :: s
         integer, intent(out) :: ierr
         ierr = 0
         call star_ptr(id,s,ierr)
         if(ierr/=0) return

         ! here is an example for adding an extra history header item
         ! also set how_many_extra_history_header_items
         ! names(1) = 'mixing_length_alpha'
         ! vals(1) = s% mixing_length_alpha

      end subroutine data_for_extra_history_header_items


      integer function how_many_extra_profile_header_items(id)
         integer, intent(in) :: id
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_profile_header_items = 0
      end function how_many_extra_profile_header_items


      subroutine data_for_extra_profile_header_items(id, n, names, vals, ierr)
         integer, intent(in) :: id, n
         character (len=maxlen_profile_column_name) :: names(n)
         real(dp) :: vals(n)
         type(star_info), pointer :: s
         integer, intent(out) :: ierr
         ierr = 0
         call star_ptr(id,s,ierr)
         if(ierr/=0) return

         ! here is an example for adding an extra profile header item
         ! also set how_many_extra_profile_header_items
         ! names(1) = 'mixing_length_alpha'
         ! vals(1) = s% mixing_length_alpha

      end subroutine data_for_extra_profile_header_items


      ! returns either keep_going or terminate.
      ! note: cannot request retry; extras_check_model can do that.
      integer function extras_finish_step(id)
         integer, intent(in) :: id
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_finish_step = keep_going

         if (s% L_nuc_burn_total .gt. 1e4) then
            has_flashed = .true.
         end if

         ! to save a profile, 
            ! s% need_to_save_profiles_now = .true.
         ! to update the star log,
            ! s% need_to_update_history_now = .true.

         ! see extras_check_model for information about custom termination codes
         ! by default, indicate where (in the code) MESA terminated
         if (extras_finish_step == terminate) s% termination_code = t_extras_finish_step
      end function extras_finish_step
      
      
      subroutine extras_after_evolve(id, ierr)
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
      end subroutine extras_after_evolve


      subroutine do1_net( &
            s, k, species, num_reactions, net_lwork, &
            n, nz, vals, ierr)
         use rates_def, only: std_reaction_Qs, std_reaction_neuQs, i_rate
         use net_def, only: Net_Info
         use net_lib, only: net_get
         use chem_def, only: chem_isos, category_name
         use eos_def, only : i_eta
         use utils_lib,only: &
            is_bad_num, realloc_double, realloc_double3
         type (star_info), pointer :: s         
         integer, intent(in) :: k, species, num_reactions, net_lwork, n, nz
         real(dp) :: vals(nz,n)
         integer, intent(out) :: ierr
 
         integer :: i, j, screening_mode
         real(dp) :: log10_rho, log10_T, alfa, beta, &
            d_eps_nuc_dRho, d_eps_nuc_dT, cat_factor
         real(dp), target :: net_work_ary(net_lwork)
         real(dp), pointer :: net_work(:)
         type (Net_Info), target :: net_info_target
         type (Net_Info), pointer :: netinfo
         
         character (len=100) :: message
         real(dp), pointer :: reaction_neuQs(:)
         integer :: sz
         real(dp) :: eps_nuc_factor
         
         logical, parameter :: dbg = .false.

         include 'formats'
         
         ierr = 0
         
         net_work => net_work_ary
         netinfo => net_info_target
      
         log10_rho = s% lnd(k)/ln10
         log10_T = s% lnT(k)/ln10

         screening_mode = get_screening_mode(s,ierr)         
         if (ierr /= 0) then
            write(*,*) 'unknown string for screening_mode: ' // trim(s% screening_mode)
            stop 'do1_net'
            return
         end if
      
         call net_get( &
            s% net_handle, .false., netinfo, species, num_reactions, s% xa(1:species,k), &
            s% T(k), log10_T, s% rho(k), log10_Rho, &
            s% abar(k), s% zbar(k), s% z2bar(k), s% ye(k), &
            s% eta(k), s% d_eos_dlnd(i_eta,k), s% d_eos_dlnT(i_eta,k), &
            s% rate_factors, s% weak_rate_factor, &
            std_reaction_Qs, std_reaction_neuQs, .false., .false., &
            s% eps_nuc(k), d_eps_nuc_dRho, d_eps_nuc_dT, s% d_epsnuc_dx(:,k), & 
            s% dxdt_nuc(:,k), s% d_dxdt_nuc_dRho(:,k), s% d_dxdt_nuc_dT(:,k), s% d_dxdt_nuc_dx(:,:,k), &
            screening_mode, s% theta_e(k), &
            s% eps_nuc_categories(:,k), &
            s% eps_nuc_neu_total(k), net_lwork, net_work, ierr)
      
         if (ierr /= 0) then
            write(*,*) 'do1_net: net_get failure for cell ', k
            return
         end if
         
         call show_stuff(s,k,net_lwork,net_work,n,nz,vals)

      end subroutine do1_net
      
      
      integer function get_screening_mode(s,ierr)
         use rates_lib, only: screening_option
         type (star_info), pointer :: s 
         integer, intent(out) :: ierr
         include 'formats'
         ierr = 0
         if (s% screening_mode_value >= 0) then
            get_screening_mode = s% screening_mode_value
            return
         end if
         get_screening_mode = screening_option(s% screening_mode, ierr)
         if (ierr /= 0) return
         s% screening_mode_value = get_screening_mode
         !write(*,2) 'get_screening_mode ' // &
         !   trim(s% screening_mode), get_screening_mode
      end function get_screening_mode
      
      
      
      subroutine show_stuff(s,k,lwork,work,n,nz,vals)
         use chem_def
         use rates_def
         use rates_lib, only: rates_reaction_id
         use net_lib, only: get_reaction_id_table_ptr, get_net_rate_ptrs
         type (star_info), pointer :: s         
         integer, intent(in) :: k, lwork, n, nz
         real(dp), pointer :: work(:)
         real(dp) :: vals(nz,n)

         integer, pointer :: reaction_id(:) ! maps net reaction number to reaction id
         integer :: i, j, ierr, species, num_reactions, rate_id
         real(dp), pointer, dimension(:) :: &
            rate_screened, rate_screened_dT, rate_screened_dRho, &
            rate_raw, rate_raw_dT, rate_raw_dRho

         include 'formats'
         
         ierr = 0
         num_reactions = s% num_reactions
         
         call get_net_rate_ptrs(s% net_handle, &
            rate_screened, rate_screened_dT, rate_screened_dRho, &
            rate_raw, rate_raw_dT, rate_raw_dRho, lwork, work, &
            ierr)
         if (ierr /= 0) then
            write(*,*) 'failed in get_net_rate_ptrs'
            stop 1
         end if
         
         call get_reaction_id_table_ptr(s% net_handle, reaction_id, ierr) 
         if (ierr /= 0) return
         
         rate_id = rates_reaction_id('r_be7_wk_li7')
         if (rate_id <= 0) then
            write(*,*) 'failed to find reaction rate id -- not valid name?'
            vals(k,1) = 0
            return
         end if
         vals(k,1) = get_rate(rate_id)
         
         rate_id = rates_reaction_id('r_li7_pa_he4')
         if (rate_id <= 0) then
            write(*,*) 'failed to find reaction rate id -- not valid name?'
            vals(k,2) = 0
            return
         end if
         vals(k,2) = get_rate(rate_id)
         
         
         contains
         
         real(dp) function get_rate(id)
            integer, intent(in) :: id
            integer :: j
            include 'formats'
            
            get_rate = -99
            do j=1,num_reactions
               if (reaction_id(j) /= rate_id) cycle
               !write(*,3) 'screened rate ' // trim(reaction_Name(reaction_id(j))), &
               !   j, k, rate_screened(j)
               if (rate_screened(j) < 1d-99) then
                  get_rate = -99
               else
                  get_rate = log10(rate_screened(j)) ! or rate_raw(j)
               end if
               return
            end do

            write(*,*) 'failed to find reaction rate id -- not in current net?'
            get_rate = -99

         end function get_rate
         
      end subroutine show_stuff

      end module run_star_extras
      
